from sklearn.preprocessing import MinMaxScaler
import numpy as np
import copy

def scale_input_data(input_data, metadata, ignore_integrated_scaler=False):
    """
        Scale input data from real values to normalized.

        Parameters
        ----------
        data : numpy.array
            Data as numpy array
        metadata : dict
            Model metadata
        ignore_integrated_scaler: bool
            If False(default), only scale data if model does not contain integrated scaler

        Returns
        -------
        numpy.array
            Scaled data
        """
    if ignore_integrated_scaler or not metadata['is_scaler_integrated']:
        return __scale_data_from_config(
            input_data,
            metadata['scaler_config']['input'],
            metadata['scaler_name']
        )
    else:
        return input_data

def scale_output_data(output_data, metadata):
    """
        Scale output data from real values to normalized.

        Parameters
        ----------
        data : numpy.array
            Data as numpy array
        metadata : dict
            Model metadata
        ignore_integrated_scaler: bool
            If False(default), only scale data if model does not contain integrated scaler

        Returns
        -------
        numpy.array
            Scaled data
        """

    if metadata['scaler_config']['output']['feature_range'] is not None:

        output_data_ranges = [
            {
                'Min' : metadata['scaler_config']['output']['fit'][0][0],
                'Max' : metadata['scaler_config']['output']['fit'][1][0]
            }
        ]

        af_range = metadata['scaler_config']['output']['feature_range']

        output_data = scale_data(output_data_ranges, af_range, output_data, False)

    return output_data

def rescale_output_data(output_data, metadata, ignore_integrated_scaler=False):
    """
        Scale data from normalized values to real values. Use after prediction.

        Parameters
        ----------
        data : numpy.array
            Data as numpy array
        metadata : dict
            Model metadata
        ignore_integrated_scaler: bool
            If False(default), only scale data if model does not contain integrated scaler

        Returns
        -------
        numpy.array
            Scaled data
        """
    if metadata['scaler_config']['output']['fit'] is not None:
        if ignore_integrated_scaler or not metadata['is_scaler_integrated']:
            recurrent_output_count = metadata.get('recurrent_output_count', None)
            config = metadata['scaler_config']['output']
            if recurrent_output_count is not None:
                config = copy.deepcopy(config)
                config['fit'][0] = config['fit'][0] * recurrent_output_count
                config['fit'][1] = config['fit'][1] * recurrent_output_count
            return __scale_data_from_config(
                output_data,
                config,
                metadata['scaler_name']
            )
    else:
        return output_data


def __scale_data_from_config(data, config, scaler_name='MinMaxScaler'):
    if scaler_name == 'MinMaxScaler':
        return __min_max_scale_data(
            data,
            config['fit'],
            config['feature_range'],
            config['inverse_transform']
        )
    else:
        raise Exception('Unknown scaler ' + scaler_name)


def __min_max_scale_data(data, fit, feature_range, inverse=False):
    scaler = MinMaxScaler(feature_range=feature_range)
    scaler.fit(fit)
    if inverse is True:
        return scaler.inverse_transform(data)
    else:
        return scaler.transform(data)


def scale_data(data_ranges, range, data, inverse=False):
    min_max_scaler = MinMaxScaler(feature_range=range)
    # set input ranges
    in_ranges = list(map(lambda x: [x['Min'], x['Max']], data_ranges))
    in_ranges = np.array(in_ranges).transpose()
    min_max_scaler.fit(in_ranges)
    if inverse is True:
        return min_max_scaler.inverse_transform(data)
    else:
        return min_max_scaler.transform(data)



from .input_extras import remove_not_used_inputs
from .scaler_extras import scale_input_data, scale_output_data
from .series_extras import pack_input_data_for_series
from .series_extras import pack_output_data_for_series
from .rnn_extras import pack_input_data_for_rnn
from .rnn_extras import pack_output_data_for_rnn
from .encode_data_set import encode_inputs, encode_output_data

def prepare_input_data(input_data, metadata):
    """
        Prepare the input for prediction or training with the following steps:
            1. Removing insignificant columns,
            2. Encoding of input features,
            3. Packing data for series,
            4. Scaling (normalizing) values,
            5. Pack data for recurrent neural network.
        Parameters
        ----------
        input_data : numpy.array
            Input data as numpy array.
        metadata : dict
            Model metadata.

        Returns
        -------
        numpy.array
            Prepared input values.
        """
    used_inputs = remove_not_used_inputs(input_data, metadata)
    if 'input_encodings' in metadata:
        used_inputs = encode_inputs(used_inputs, metadata['input_encodings'])
    if metadata['has_rolling']:
        used_inputs = pack_input_data_for_series(used_inputs, metadata)
    used_inputs = scale_input_data(used_inputs, metadata)
    if metadata.get('recurrent_input_count', None) is not None and metadata.get('recurrent_output_count', None) is not None:
        used_inputs = pack_input_data_for_rnn(used_inputs, metadata)
    return used_inputs


def prepare_output_data(output_data, metadata):
    """
        Prepare the output for prediction or training with the following steps:
            1. Encode output data,
            2. Packing data for series,
            3. Scale output data,
            4. Pack data for recurrent neural network.
        Parameters
        ----------
        output_data : numpy.array
            Output data as numpy array.
        metadata : dict
            Model metadata.

        Returns
        -------
        numpy.array
            Prepared output values.
        """
    if 'output_encoding' in metadata:
        output_data = encode_output_data(output_data, metadata)
    if metadata['has_rolling']:
        output_data = pack_output_data_for_series(output_data, metadata)
    output_data = scale_output_data(output_data, metadata)
    if metadata.get('recurrent_input_count', None) is not None and metadata.get('recurrent_output_count', None) is not None:
        output_data = pack_output_data_for_rnn(output_data, metadata)
    return output_data

def pack_output_data_for_comparison_with_predictions(output_data, metadata):
    """
        Prepare the output for comparison with the following steps:
            1. Encode output data,
            2. Packing data for series or pack data for recurrent neural network.
        Parameters
        ----------
        output_data : numpy.array
            Output data as numpy array.
        metadata : dict
            Model metadata.

        Returns
        -------
        numpy.array
            Prepared output values.
        """
    if 'output_encodings' in metadata:
        output_data = encode_output_data(output_data, metadata['output_encodings'])
    if metadata['has_rolling']:
        output_data = pack_output_data_for_series(output_data, metadata)
    if metadata.get('recurrent_input_count', None) is not None and metadata.get('recurrent_output_count', None) is not None:
        output_data = pack_output_data_for_rnn(output_data, metadata)
    return output_data


# coding: utf-8

# flake8: noqa

from __future__ import absolute_import
from .series_extras import pack_data_for_series, get_series_padding, pack_input_data_for_series, pack_output_data_for_series
from .rnn_extras import pack_input_data_for_rnn, pack_output_data_for_rnn
from .scaler_extras import scale_input_data, scale_output_data, rescale_output_data
from .input_extras import remove_not_used_inputs
from .data_extras import prepare_input_data, prepare_output_data
from .encode_data_set import encode_inputs, encode_output_data, decode_output_data
from .load_models import load_model


import numpy as np
from .encode_variable import VariableEncoding

def encode_inputs(input_data_set, input_encoding_info):
    """
        Test set input data encoding. Use before prediction.

        Parameters
        ----------
        input_data_set : numpy.array
            Data as numpy array
        input_encoding_info : [dict]
            Variables encoding info from training input data

        Returns
        -------
        numpy.array
            Encoded input data
    """
    input_data_set = np.array(input_data_set)

    i = 0
    enc_index = 0
    number_of_variables = len(input_encoding_info)
    while i < number_of_variables:
        variable_encoder_info = input_encoding_info[enc_index]

        if variable_encoder_info['type'] != 'None':
            variable = input_data_set[:, i:i+1]
            encoded_variable = VariableEncoding.encode(variable, variable_encoder_info) # encode each variable as in training
            
            if i == 0:
                input_data_set = np.concatenate((encoded_variable, input_data_set[:, i+1:]), axis=1)
            elif i == number_of_variables-1:
                input_data_set = np.concatenate((input_data_set[:, :i], encoded_variable), axis=1)
            else:
                input_data_set = np.concatenate((input_data_set[:, :i], encoded_variable, input_data_set[:, i+1:]), axis=1)

            if variable.shape[1] != encoded_variable.shape[1]:
                number_of_variables = number_of_variables - variable.shape[1] + encoded_variable.shape[1]
                i = i - variable.shape[1] + encoded_variable.shape[1]
        i += 1
        enc_index += 1
    
    return input_data_set

def encode_output_data(output_data_set, metadata):
    """
        Test set output data encoding.

        Parameters
        ----------
        output_data_set : numpy.array
            Test set output data as numpy array
        metadata : [dict]
            Optimization metadata

        Returns
        -------
        numpy.array
            Test set encoded output data
    """
    

    if 'output_encoding' in metadata:
        output_data_set = np.array(output_data_set)
        output_data_set = VariableEncoding.encode(output_data_set, metadata['output_encoding'][0])
    else:
        raise Exception ("The output variable has not been encoded so this method cannot be applied.") 

    
    return output_data_set

def decode_output_data(output_data_set, metadata, threshold=0.5):
    """
        Test set output data decoding. Use after prediction.

        Parameters
        ----------
        output_data_set : numpy.array
            Predicted data as numpy array
        metadata : [dict]
            Optimization metadata

        Returns
        -------
        numpy.array
            Decoded predicted output data
    """

    if 'output_encoding' in metadata:
        output_data_set = VariableEncoding.decode(output_data_set, metadata['output_encoding'][0], threshold)
    else:
        raise Exception ("The output variable has not been encoded so this method cannot be applied.") 

    return output_data_set


import numpy as np

def get_series_padding(metadata):
    """
        Get left and right data padding

        Parameters
        ----------
        metadata : dict
            Model metadata

        Returns
        -------
        (int, int)
            Left padding, right padding
    """
    input_count = len(metadata['input_shifts'])
    output_count = len(metadata['output_shifts'])

    in_range = range(input_count)
    out_range = range(output_count)

    left_input = max(list(map(
        lambda x: -metadata['input_shifts'][x] + metadata['input_windows'][x], in_range)))
    left_output = max(list(map(
        lambda x: -metadata['output_shifts'][x] + metadata['output_windows'][x], out_range)))

    right_input = max(
        list(map(lambda x: metadata['input_shifts'][x], in_range)))
    right_output = max(
        list(map(lambda x: metadata['output_shifts'][x], out_range)))

    sample_step = metadata['output_sample_step']
    left_padding = max(0, left_input-1, left_output-1)
    left_padding = int(sample_step * np.ceil(left_padding / sample_step))
    right_padding = max(right_input, right_output, 0)

    return left_padding, right_padding


def pack_input_data_for_series(input_data, metadata):
    """
        Packing input data for series

        Parameters
        ----------
        data : numpy.array
            Data as numpy array
        metadata : dict
            Model metadata

        Returns
        -------
        numpy.array
            Packed data
    """
    input_count = len(metadata['input_shifts'])
    sample_step = metadata['output_sample_step']

    left_padding, right_padding = get_series_padding(metadata)

    x_data = []
    rows_count = input_data.shape[0]-right_padding
    for i in range(left_padding, rows_count, sample_step):
        x_row = __get_input_row(input_data, metadata, i, input_count)
        x_data.append(x_row)

    x_data = np.array(x_data)
    return x_data


def pack_output_data_for_series(output_data, metadata):
    """
        Packing output data for series

        Parameters
        ----------
        data : numpy.array
            Data as numpy array
        metadata : dict
            Model metadata

        Returns
        -------
        numpy.array
            Packed data
    """
    output_count = len(metadata['output_shifts'])
    sample_step = metadata['output_sample_step']

    left_padding, right_padding = get_series_padding(metadata)

    y_data = []
    rows_count = output_data.shape[0]-right_padding
    for i in range(left_padding, rows_count, sample_step):
        y_row = __get_output_row(output_data, metadata, i, output_count, 0)
        y_data.append(y_row)

    y_data = np.array(y_data)
    return y_data


def pack_data_for_series(data, metadata):
    """
        Packing input and output data for series

        Parameters
        ----------
        data : numpy.array
            Data as numpy array
        metadata : dict
            Model metadata

        Returns
        -------
        (numpy.array, numpy.array)
            Tuple where first value is input data(x) and second is output data(y)
    """
    input_count = len(metadata['input_shifts'])
    output_count = len(metadata['output_shifts'])
    sample_step = metadata['output_sample_step']

    left_padding, right_padding = get_series_padding(metadata)

    x_data = []
    y_data = []
    rows_count = data.shape[0]-right_padding
    for i in range(left_padding, rows_count, sample_step):
        
        x_row = __get_input_row(data, metadata, i, input_count)
        x_data.append(x_row)

        y_row = __get_output_row(data, metadata, i, output_count, input_count)
        y_data.append(y_row)

    x_data = np.array(x_data)
    y_data = np.array(y_data)
    return (x_data, y_data)


def __get_input_row(data, metadata, i, input_count):
    x_row = []
    for j in range(input_count):
        shift = metadata['input_shifts'][j]
        window = metadata['input_windows'][j]
        step = metadata['input_steps'][j]
        aggregation_type = metadata['input_aggregation_types'][j]
        if step is None:
            step = window if aggregation_type != 'none' else 1
        offset = i+1+shift
        d = data[offset-window:offset, j]
        n = len(d)
        if aggregation_type == 'sum':
            r = range((n) % step, n, step)
            new_d = [sum(d[s:s+step]) for s in r]
        elif aggregation_type == 'avg':
            r = range((n) % step, n, step)
            new_d = [np.mean(d[s:s+step]) for s in r]
        else:
            r = range((n-1) % step, n, step)
            new_d = [d[s] for s in r]

        x_row = np.concatenate([x_row, new_d])
    return x_row


def __get_output_row(data, metadata, i, output_count, input_offset=0):
    y_row = []
    for j in range(output_count):
        c = input_offset + j
        shift = metadata['output_shifts'][j]
        window = metadata['output_windows'][j]
        offset = i+1+shift
        d = data[offset-window:offset,c]
        y_row = np.concatenate([y_row, d])

    return y_row



# coding: utf-8

"""
    BlackFoxExtras

    BlackFoxExtras

"""


from setuptools import setup, find_packages  # noqa: H301

NAME = "blackfox_extras"
VERSION = "1.0.0"

REQUIRES = [
    "blackfox >= 6.1.0",
    "numpy >= 1.19.1",
    "scikit-learn >= 1.0.0",
    "pandas >= 0.25.3",
    "category_encoders >= 2.2.2",
    "tensorflow >= 2.6.1",
    "xgboost == 1.5.0"
    ]

setup(
    name=NAME,
    version=VERSION,
    description="BlackFox Extras",
    author="Tomislav Mrdja",
    author_email="",
    url="https://github.com/vodena/BlackFoxPythonExtras",
    keywords=["BlackFox"],
    install_requires=REQUIRES,
    packages=find_packages(exclude=["test.*", "test"]),
    include_package_data=True,
    long_description="""\
        BlackFox Extras
    """
)



from sklearn.preprocessing import OneHotEncoder, LabelEncoder
from category_encoders import TargetEncoder
from category_encoders.sum_coding import SumEncoder
import pandas as pd
import numpy as np

class VariableEncoding:
    @staticmethod
    def get_available_encoder_types():
        return ["None", "OneHot", "Dummy", "Target", "Effect", "CountOfFrequency", "OrderInteger", "Label"]

    @staticmethod
    def get_available_decoder_types():
        return ["OneHot", "Dummy", "Label"]

    @staticmethod
    def encode(variable, variable_encoder_info):
        """
            Encoding each variable of test set as it was done in the training set.

            Parameters
            ----------
            data : numpy.array
                Data as numpy array.
            encoding_info : dict
                Variable encoding info from training data.

            Returns
            -------
            numpy.array
                Encoded variable.
        """
        variable = np.array(variable).reshape(len(variable), 1)
        encoding_type = variable_encoder_info['type']
        if encoding_type in VariableEncoding.get_available_encoder_types():
            variable_str = []
            for elem in variable:
                if type(elem) is not str:
                    variable_str.append(str(elem[0]))
                else:
                    variable_str.append(elem)
            variable_str = np.array(variable_str).reshape(len(variable_str), 1)
            #variable_str.dtype = str

            if encoding_type == "OneHot":
                training_input_categories = np.array(variable_encoder_info['categories'])           
                enc = OneHotEncoder(categories = [training_input_categories])
                enc.fit(training_input_categories.reshape(len(training_input_categories), 1))
                encoded_variable = enc.transform(variable_str).toarray()
            elif encoding_type == "Dummy":
                training_input_categories = np.array(variable_encoder_info['categories'])           
                enc = OneHotEncoder(categories = [training_input_categories], drop = 'first')
                enc.fit(training_input_categories.reshape(len(training_input_categories), 1))
                encoded_variable = enc.transform(variable_str).toarray()
            elif encoding_type == "Target":
                training_input_mapping = variable_encoder_info['mapping']
                variable_str = pd.Series(variable_str.reshape(len(variable_str),))
                encoded_variable = variable_str.map(training_input_mapping)
                encoded_variable = np.array(encoded_variable).reshape(len(encoded_variable), 1)
            elif encoding_type == "Effect":
                variable_str = list(variable_str.reshape(len(variable_str),))
                encoded_variable = np.array([variable_encoder_info['mapping'][elem] for elem in variable_str])
            elif encoding_type == "CountOfFrequency":
                variable_str = pd.Series(variable_str.reshape(len(variable_str), ))
                encoded_variable = variable_str.map(variable_encoder_info['mapping'])
                encoded_variable = np.array(encoded_variable).reshape(len(encoded_variable), 1)
            elif encoding_type == "Label":
                training_input_categories = np.array(variable_encoder_info['categories'])
                enc = LabelEncoder()
                enc.classes_ = training_input_categories
                enc.fit(training_input_categories.reshape(len(training_input_categories), 1))
                encoded_variable = enc.transform(variable_str)
                encoded_variable = np.array(encoded_variable).reshape(len(encoded_variable), 1)
            else:
                variable_str = pd.Series(variable_str.reshape(len(variable_str), ))
                encoded_variable = variable_str.map(variable_encoder_info['mapping'])
                encoded_variable = np.array(encoded_variable).reshape(len(encoded_variable), 1)
                
            return encoded_variable
        else:
            raise Exception ("Encoding type " + str(variable_encoder_info['type']) + " is not suported.")
    
    

    @staticmethod
    def decode(variable, variable_encoder_info, threshold):
        """
            Decode variable in relation to the encode method used in the optimization process.

            Parameters
            ----------
            variable : numpy.array
                Data as numpy array.
            variable_encoder_info : dict
                Variable encoding info from training input data.
            variable_encoder_info : float
                The threshold that determines the binary classes.

            Returns
            -------
            numpy.array
                Decoded variable.
        """

        decoding_type = variable_encoder_info['type']

        if decoding_type in VariableEncoding.get_available_decoder_types():
            if decoding_type == "OneHot":
                predicted_var = np.array(variable)
                predicted_var =  np.where(predicted_var == predicted_var.max(axis=1).reshape(len(predicted_var),1), 1, 0)

                training_input_categories = np.array(variable_encoder_info['categories'])           
                enc = OneHotEncoder(categories = [training_input_categories])
                enc.fit(training_input_categories.reshape(len(training_input_categories), 1))
                decoded_variable = enc.inverse_transform(predicted_var)
                decoded_variable = np.array([elem[0] for elem in decoded_variable]).reshape(len(decoded_variable), 1)
            elif decoding_type == "Label":
                predicted_var = np.array(variable)

                training_input_categories = np.array(variable_encoder_info['categories'])
                enc = LabelEncoder()
                enc.classes_ = training_input_categories
                enc.fit(training_input_categories.reshape(len(training_input_categories), 1))
                decoded_variable = enc.inverse_transform(predicted_var)
                decoded_variable = np.array([elem for elem in decoded_variable]).reshape(len(decoded_variable), 1)
            else:
                predicted_var = np.array(variable).reshape(len(variable), 1)
                predicted_var =  np.where(predicted_var > threshold, 1, 0)

                training_input_categories = np.array(variable_encoder_info['categories'])  
                enc = OneHotEncoder(categories = [training_input_categories], drop = 'first')
                enc.fit(training_input_categories.reshape(len(training_input_categories), 1))
                decoded_variable = enc.inverse_transform(predicted_var)
            try:
                decoded_variable = np.array([int(float(elem[0])) for elem in decoded_variable]).reshape(len(decoded_variable), 1)
                return decoded_variable
            except:
                return decoded_variable

        else:
            raise Exception ("Decoding type " + str(variable_encoder_info['type']) + " is not suported.")